# coding=utf-8
import tensorflow as tf


def le_conv(x, edge_index, edge_weight,
            self_kernel, self_bias,
            aggr_self_kernel, aggr_self_bias,
            aggr_neighbor_kernel, aggr_neighbor_bias, activation=None):
    """
    Functional API for LeConv in ASAP.

    h_i = activation(x_i @ self_kernel + \sum_{j} (x_i @ aggr_self_kernel - x_j @ aggr_neighbor_kernel))

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param self_kernel: Please look at the formula above.
    :param aggr_self_kernel: Please look at the formula above.
    :param aggr_neighbor_kernel: Please look at the formula above.
    :param activation: Activation function to use.
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    if edge_weight is None:
        edge_weight = tf.ones([tf.shape(edge_index)[1]], dtype=tf.float32)

    num_nodes = tf.shape(x)[0]
    self_h = x @ self_kernel
    if self_bias is not None:
        self_h += self_bias

    aggr_self_h = x @ aggr_self_kernel
    if aggr_self_bias is not None:
        aggr_self_h += aggr_self_bias

    aggr_neighbor_h = x @ aggr_neighbor_kernel
    if aggr_neighbor_bias is not None:
        aggr_neighbor_h += aggr_neighbor_bias

    row, col = edge_index[0], edge_index[1]

    repeated_aggr_self_h = tf.gather(aggr_self_h, col)
    repeated_aggr_neighbor_h = tf.gather(aggr_neighbor_h, col)
    repeated_aggr_h = (repeated_aggr_self_h - repeated_aggr_neighbor_h) * tf.expand_dims(edge_weight, axis=-1)
    aggr_h = tf.math.unsorted_segment_sum(repeated_aggr_h, row, num_nodes)

    h = self_h + aggr_h

    if activation is not None:
        h = activation(h)

    return h






# coding=utf-8

import warnings

warnings.simplefilter('always', DeprecationWarning)

import tensorflow as tf

if tf.__version__[0] == "1":
    tf.enable_eager_execution()

from . import nn, utils, data, datasets, layers
from .data.graph import Graph, BatchGraph, HeteroDictGraph
# from .sparse.sparse_adj import SparseAdj



# coding=utf-8
import tf_sparse as tfs

from tf_geometric.nn.conv.gcn import gcn, gcn_build_cache_for_graph, gcn_build_cache_by_adj
import tensorflow as tf
import warnings
import numpy as np

from tf_geometric.utils.tf_sparse_utils import compute_num_or_size_splits


class GCN(tf.keras.Model):
    """
    Graph Convolutional Layer
    """

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        if self.num_splits is not None:
            num_h_features = self.units if self.use_kernel else num_features
            self.num_or_size_splits = compute_num_or_size_splits(num_h_features, self.num_splits)

        if self.use_kernel:
            self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                          initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units if self.use_kernel else num_features],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def __init__(self, units, activation=None,
                 use_kernel=True,
                 use_bias=True,
                 norm="both", add_self_loop=True, sym=True,
                 renorm=True, improved=False,
                 edge_drop_rate=0.0,
                 num_splits=None,
                 num_or_size_splits=None,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param norm: normalization mode both|left|right:
            - both: (D^(-1/2)A)D^(-1/2);
            - left: D^(-1/2)A; 
            - right: AD^(-1/2);
        :param add_self_loop: Whether add self-loop to adj during normalization.
        :param sym: Optional, only used when norm=="both". Setting sym=True indicates that the input
            sparse_adj is symmetric.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param edge_drop_rate: Dropout rate of the propagation weights.
        :param num_or_size_splits: Split (XW) to compute A(XW) for large graphs (Not affecting the output).
            See the num_or_size_splits param of the tf.split API.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        self.activation = activation
        self.use_kernel = use_kernel
        self.use_bias = use_bias

        self.edge_drop_rate = edge_drop_rate

        self.kernel = None
        self.bias = None

        self.norm = norm
        self.add_self_loop = add_self_loop
        self.sym = sym
        self.renorm = renorm
        self.improved = improved

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        if num_splits is not None and num_or_size_splits is not None:
            raise Exception("cannot provide both num_splits and num_or_size_splits for GCN")

        self.num_splits = num_splits
        self.num_or_size_splits = num_or_size_splits

    def build_cache_by_adj(self, sparse_adj, override=False, cache=None):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        return gcn_build_cache_by_adj(sparse_adj,
                                      self.norm, self.add_self_loop, self.sym,
                                      self.renorm, self.improved, override=override, cache=cache)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph,
                                  self.norm, self.add_self_loop, self.sym,
                                  self.renorm, self.improved, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn("'GCN.cache_normed_edge(graph, override)' is deprecated, use 'GCN.build_cache_for_graph(graph, override)' instead", DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, sparse_adj], [x, edge_index], 
            or [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if isinstance(inputs[1], tfs.SparseMatrix):
            x, sparse_adj = inputs
        elif len(inputs) == 3:
            x, edge_index, edge_weight = inputs
            num_nodes = tfs.shape(x)[0]
            sparse_adj = tfs.SparseMatrix(edge_index, value=edge_weight, shape=[num_nodes, num_nodes])
        elif len(inputs) == 2:
            x, edge_index = inputs
            num_nodes = tfs.shape(x)[0]
            sparse_adj = tfs.SparseMatrix(edge_index, shape=[num_nodes, num_nodes])

        return gcn(x, sparse_adj, self.kernel, self.bias,
                   activation=self.activation,
                   norm=self.norm, add_self_loop=self.add_self_loop, sym=self.sym,
                   renorm=self.renorm, improved=self.improved,
                   edge_drop_rate=self.edge_drop_rate,
                   num_or_size_splits=self.num_or_size_splits,
                   training=training, cache=cache)



# coding=utf-8

import numpy as np
import os

import networkx as nx
from tf_geometric.data.graph import Graph
from tf_geometric.data.dataset import DownloadableDataset
import json
import scipy.sparse as sp
from tf_geometric.utils.data_utils import load_cache
from tf_geometric.utils.graph_utils import convert_edge_to_directed


class _BaseRedditDataset(DownloadableDataset):

    def __init__(self, dataset_root_path=None, cache_name=None):
        super().__init__(dataset_name="reddit",
                         download_urls=[
                             "https://data.dgl.ai/dataset/reddit.zip"
                         ],
                         download_file_name="reddit.zip",
                         cache_name=cache_name,
                         dataset_root_path=dataset_root_path,
                         )

    def process(self):

        common_data_path = os.path.join(self.raw_root_path, "reddit_data.npz")
        common_data = np.load(common_data_path)

        x = common_data["feature"]
        y = common_data["label"]

        mask = common_data["node_types"]
        full_index = np.arange(len(x), dtype=np.int32)
        train_index = full_index[mask == 1]
        valid_index = full_index[mask == 2]
        test_index = full_index[mask == 3]

        graph_data_path = os.path.join(self.raw_root_path, "reddit_graph.npz")

        adj = sp.load_npz(graph_data_path)
        edge_index = np.stack([adj.row, adj.col], axis=0)

        graph = Graph(x=x, y=y, edge_index=edge_index)
        return graph, (train_index, valid_index, test_index)


class TransductiveRedditDataset(_BaseRedditDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_root_path=dataset_root_path, cache_name="transductive_cache.p")


class InductiveRedditDataset(_BaseRedditDataset):

    def __init__(self, dataset_root_path=None):
        super().__init__(dataset_root_path=dataset_root_path, cache_name="inductive_cache.p")

    def process(self):
        graph, (train_index, valid_index, test_index) = super().process()
        train_graph = graph.sample_new_graph_by_node_index(train_index)
        valid_graph = graph.sample_new_graph_by_node_index(valid_index)
        test_graph = graph.sample_new_graph_by_node_index(test_index)
        return train_graph, valid_graph, test_graph















# coding=utf-8


import tensorflow as tf

from tf_geometric.nn.pool.diff_pool import diff_pool


class DiffPool(tf.keras.Model):
    """
    OOP API for DiffPool: "Hierarchical graph representation learning with differentiable pooling"
    """

    def __init__(self, feature_gnn, assign_gnn, units, num_clusters, activation=None, use_bias=True,
                 bias_regularizer=None, *args, **kwargs):
        """
        DiffPool

        :param feature_gnn: A GNN model to learn pooled node features, [x, edge_index, edge_weight] => updated_x,
            where updated_x corresponds to high-order node features.
        :param assign_gnn: A GNN model to learn cluster assignment for the pooling, [x, edge_index, edge_weight] => updated_x,
            where updated_x corresponds to the cluster assignment matrix.
        :param units: Positive integer, dimensionality of the output space. It must be provided if you set use_bias=True.
        :param num_clusters: Number of clusters for pooling.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector. If true, the "units" parameter must be provided.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.feature_gnn = feature_gnn
        self.assign_gnn = assign_gnn

        self.num_clusters = num_clusters
        self.activation = activation

        if use_bias and units is None:
            raise Exception("The \"units\" parameter is required when you set use_bias=True.")

        if use_bias:
            self.bias = self.add_weight("bias", shape=[units],
                                        initializer="zeros", regularizer=bias_regularizer)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight, node_graph_index]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Pooled graph: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
        """
        x, edge_index, edge_weight, node_graph_index = inputs

        return diff_pool(x, edge_index, edge_weight, node_graph_index,
                         self.feature_gnn, self.assign_gnn, self.num_clusters,
                         bias=self.bias, activation=self.activation, training=training, cache=cache)



# coding=utf-8

import tensorflow as tf
from tf_geometric.nn.pool.set2set import set2set


class Set2Set(tf.keras.Model):
    """
    OOP API for Set2Set
    """

    def __init__(self, num_iterations=4, *args, **kwargs):
        """
        Set2Set

        :param num_iterations: Number of iterations for attention.
        """
        super().__init__(*args, **kwargs)
        self.num_iterations = num_iterations
        self.lstm = None

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.lstm = tf.keras.layers.LSTM(num_features, return_sequences=True, return_state=True)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, node_graph_index]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Graph features, shape: [num_graphs, num_node_features * 2]
        """
        x, node_graph_index = inputs

        return set2set(x, node_graph_index, self.lstm, self.num_iterations, training=training)



# coding=utf-8
import os.path
import json
from collections import defaultdict
import numpy as np
from scipy.io import loadmat

from tf_geometric.data.graph import HeteroDictGraph
from tf_geometric.data.dataset import DownloadableDataset


class _NARSAcademicDataset(DownloadableDataset):

    def __init__(self, dataset_name, dataset_root_path=None):
        """

        :param dataset_name: "nars_academic_acm"
        :param dataset_root_path:
        """
        self.sub_dataset_name = dataset_name.split("_")[-1]

        super().__init__(dataset_name=dataset_name,
                         download_urls=[
                             "https://github.com/CrawlScript/gnn_datasets/raw/master/nars_academic/{}.zip".format(self.sub_dataset_name)
                         ],
                         download_file_name="{}.zip".format(self.sub_dataset_name),
                         cache_name=None,
                         dataset_root_path=dataset_root_path,
                         )

    # https://github.com/pyg-team/pytorch_geometric/blob/master/torch_geometric/datasets/hgb_dataset.py
    def process(self):
        from scipy import io as sio

        data = loadmat(os.path.join(self.raw_root_path, "acm.mat"))
        p_vs_l = data['PvsL']  # paper-field?
        p_vs_a = data['PvsA']  # paper-author
        p_vs_t = data['PvsT']  # paper-term, bag of words
        p_vs_c = data['PvsC']  # paper-conference, labels come from that

        # We assign
        # (1) KDD papers as class 0 (data mining),
        # (2) SIGMOD and VLDB papers as class 1 (database),
        # (3) SIGCOMM and MOBICOMM papers as class 2 (communication)
        conf_ids = [0, 1, 9, 10, 13]
        label_ids = [0, 1, 2, 2, 1]

        p_vs_c_filter = p_vs_c[:, conf_ids]
        p_selected = (p_vs_c_filter.sum(1) != 0).A1.nonzero()[0]
        p_vs_l = p_vs_l[p_selected]
        p_vs_a = p_vs_a[p_selected]
        p_vs_t = p_vs_t[p_selected]
        p_vs_c = p_vs_c[p_selected]

        # pa = dgl.bipartite(p_vs_a, 'paper', 'pa', 'author')
        # pl = dgl.bipartite(p_vs_l, 'paper', 'pf', 'field')

        p_vs_a = p_vs_a.tocoo()
        p_vs_l = p_vs_l.tocoo()

        # pa = dgl.heterograph({('paper', 'pa', 'author'): (p_vs_a.col, p_vs_a.row)})
        # pl = dgl.heterograph({('paper', 'pf', 'field'): (p_vs_l.col, p_vs_l.row)})

        # gs = [pa, pl]
        # hg = dgl.hetero_from_relations(gs)
        # hg = dgl.heterograph({
        #     ('paper', 'pa', 'author'): (p_vs_a.row, p_vs_a.col),
        #     ('paper', 'pf', 'field'): (p_vs_l.row, p_vs_l.col)
        # })



        edge_index_dict = {
            ('paper', 'pa', 'author'): np.stack([p_vs_a.row, p_vs_a.col], axis=0).astype(np.int64),
            ('paper', 'pf', 'field'): np.stack([p_vs_l.row, p_vs_l.col], axis=0).astype(np.int64)
        }

        # features = torch.FloatTensor(p_vs_t.toarray())
        # features = p_vs_t.toarray().astype(np.float64)

        num_authors = p_vs_a.col.max() + 1
        num_fields = p_vs_l.col.max() + 1

        x_dict = {
            "paper": p_vs_t.toarray().astype(np.float64),
            "author": np.zeros([num_authors, 1], dtype=np.float32),
            "field": np.zeros([num_fields, 1], dtype=np.float32),
        }

        pc_p, pc_c = p_vs_c.nonzero()
        labels = np.zeros(len(p_selected), dtype=np.int64)
        for conf_id, label_id in zip(conf_ids, label_ids):
            labels[pc_p[pc_c == conf_id]] = label_id

        # labels = torch.LongTensor(labels)
        labels = np.array(labels, dtype=np.int64)

        y_dict = {"paper": labels}

        # num_classes = 3

        float_mask = np.zeros(len(pc_p))
        for conf_id in conf_ids:
            pc_c_mask = (pc_c == conf_id)
            float_mask[pc_c_mask] = np.random.permutation(np.linspace(0, 1, pc_c_mask.sum()))
        train_index = np.where(float_mask <= 0.2)[0]
        valid_index = np.where((float_mask > 0.2) & (float_mask <= 0.3))[0]
        test_index = np.where(float_mask > 0.3)[0]

        # hg.nodes["paper"].data["feat"] = features

        hetero_graph = HeteroDictGraph(x_dict=x_dict, edge_index_dict=edge_index_dict, y_dict=y_dict)

        target_node_type = "paper"

        return hetero_graph, target_node_type, (train_index, valid_index, test_index)

        # return hg, labels, num_classes, train_idx, val_idx, test_idx


class NARSACMDataset(_NARSAcademicDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("nars_academic_acm", dataset_root_path)




# coding=utf-8
from tf_geometric.datasets.ppi import PPIDataset
from tf_geometric.datasets.tu import TUDataset
from tf_geometric.datasets.planetoid import PlanetoidDataset, CoraDataset, CiteseerDataset, PubmedDataset, \
    SupervisedCoraDataset, SupervisedCiteseerDataset, SupervisedPubmedDataset
from tf_geometric.datasets.blog_catalog import MultiLabelBlogCatalogDataset
from tf_geometric.datasets.reddit import TransductiveRedditDataset, InductiveRedditDataset
from tf_geometric.datasets.ogb import OGBNodePropPredDataset
from tf_geometric.datasets.model_net import ModelNet10Dataset, ModelNet40Dataset
from tf_geometric.datasets.csr_npz import CSRNPZDataset
from tf_geometric.datasets.amazon_electronics import AmazonElectronicsDataset, AmazonComputersDataset, \
    AmazonPhotoDataset
from tf_geometric.datasets.coauthor import CoauthorDataset, CoauthorCSDataset, CoauthorPhysicsDataset
from tf_geometric.datasets.abnormal import FDAmazonDataset, FDYelpChiDataset
from tf_geometric.datasets.hgb import HGBDataset, HGBACMDataset, HGBDBLPDataset, HGBFreebaseDataset, HGBIMDBDataset
from tf_geometric.datasets.nars_academic import NARSACMDataset




# coding=utf-8
import tensorflow as tf

from tf_geometric.nn.pool.cluster_pool import cluster_pool
from tf_geometric.utils.graph_utils import convert_dense_assign_to_edge


def diff_pool_coarsen(x, edge_index, edge_weight, node_graph_index, dense_assign,
                      num_nodes=None, num_clusters=None, num_graphs=None):
    """
    Coarsening method for DiffPool.
    Coarsen the input BatchGraph (graphs) based on cluster assignment of nodes and output pooled BatchGraph (graphs).
    Graphs should be modeled as a BatchGraph like format and each graph has the same number of clusters.

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param dense_assign: Tensor, [num_nodes, num_clusters], cluster assignment matrix of nodes.
    :param num_nodes: Number of nodes, Optional, used for boosting performance.
    :param num_clusters: Number of clusters, Optional, used for boosting performance.
    :param num_graphs: Number of graphs, Optional, used for boosting performance.
    :return: Pooled BatchGraph: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
    """

    # Manually passing in num_nodes, num_clusters, and num_graphs can boost the performance
    dense_shape = tf.shape(dense_assign)

    if num_nodes is None:
        num_nodes = dense_shape[0]

    if num_clusters is None:
        num_clusters = dense_shape[1]

    if num_graphs is None:
        num_graphs = tf.reduce_max(node_graph_index) + 1

    if edge_weight is None:
        num_edges = tf.shape(edge_index)[-1]
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    assign_edge_index, assign_edge_weight = convert_dense_assign_to_edge(dense_assign, node_graph_index,
                                                                         num_nodes=num_nodes, num_clusters=num_clusters)

    # Coarsen in a large BatchGraph.
    # Here num_clusters is the sum of number of clusters of graphs in the BatchGraph
    pooled_x, pooled_edge_index, pooled_edge_weight = cluster_pool(x, edge_index, edge_weight, assign_edge_index, assign_edge_weight,
                                                                   num_clusters=num_clusters * num_graphs, num_nodes=num_nodes)

    pooled_node_graph_index = tf.reshape(tf.tile(tf.expand_dims(tf.range(num_graphs), axis=-1), [1, num_clusters]), -1)

    return pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index


def diff_pool(x, edge_index, edge_weight, node_graph_index,
              feature_gnn, assign_gnn,
              num_clusters, bias=None, activation=None, cache=None, training=None):
    """
    Functional API for DiffPool: "Hierarchical graph representation learning with differentiable pooling"

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param feature_gnn: A GNN model to learn pooled node features, [x, edge_index, edge_weight] => updated_x,
        where updated_x corresponds to high-order node features.
    :param assign_gnn: A GNN model to learn cluster assignment for the pooling, [x, edge_index, edge_weight] => updated_x,
        where updated_x corresponds to the cluster assignment matrix.
    :param num_clusters: Number of clusters for pooling.
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: [pooled_x, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index]
    """

    if edge_weight is None:
        num_edges = tf.shape(edge_index)[-1]
        edge_weight = tf.ones([num_edges], dtype=tf.float32)

    num_nodes = tf.shape(x)[0]
    num_graphs = tf.reduce_max(node_graph_index) + 1

    if cache is None:
        assign_logits = assign_gnn([x, edge_index, edge_weight], training=training)
        h = feature_gnn([x, edge_index, edge_weight], training=training)
    else:
        assign_logits = assign_gnn([x, edge_index, edge_weight], training=training, cache=cache)
        h = feature_gnn([x, edge_index, edge_weight], training=training, cache=cache)

    assign_probs = tf.nn.softmax(assign_logits, axis=-1)

    pooled_h, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index = diff_pool_coarsen(
        h, edge_index, edge_weight, node_graph_index, assign_probs,
        num_nodes=num_nodes, num_clusters=num_clusters, num_graphs=num_graphs
    )

    if bias is not None:
        pooled_h += bias

    if activation is not None:
        pooled_h = activation(pooled_h)

    return pooled_h, pooled_edge_index, pooled_edge_weight, pooled_node_graph_index


# coding=utf-8

import tensorflow as tf
import tf_sparse as tfs
import numpy as np

def sparse_tensor_gather_sub(sparse_tensor, sub_index, axis=0):
    gather_index = sparse_tensor.indices[:, axis]
    if axis in [0, -2]:
        other_axis = 1
    else:
        other_axis = 0

    dense_shape = tf.shape(sparse_tensor)

    index_mask = tf.scatter_nd(tf.expand_dims(sub_index, axis=-1), tf.ones_like(sub_index), [dense_shape[axis]])
    index_mask = tf.cast(index_mask, tf.bool)

    gather_mask = tf.gather(index_mask, gather_index)

    masked_values = tf.boolean_mask(sparse_tensor.values, gather_mask)
    masked_indices = tf.boolean_mask(sparse_tensor.indices, gather_mask)

    reverse_index = tf.cast(tf.fill([tf.reduce_max(sub_index) + 1], -1), tf.int32)
    reverse_index = tf.tensor_scatter_nd_update(reverse_index, tf.expand_dims(sub_index, axis=-1),
                                                tf.range(tf.shape(sub_index)[0]))

    masked_gather_index = masked_indices[:, axis]
    masked_gather_index = tf.gather(reverse_index, masked_gather_index)
    masked_gather_index = tf.cast(masked_gather_index, dtype=tf.int64)

    masked_other_index = masked_indices[:, other_axis]

    new_indices = [None, None]
    new_indices[axis] = masked_gather_index
    new_indices[other_axis] = masked_other_index

    new_shape = [None, None]
    new_shape[axis] = tf.shape(sub_index)[0]
    new_shape[other_axis] = dense_shape[other_axis]
    new_shape = tf.cast(new_shape, tf.int64)

    new_indices = tf.stack(new_indices, axis=1)

    new_sparse_tensor = tf.sparse.SparseTensor(
        indices=new_indices,
        values=masked_values,
        dense_shape=new_shape
    )
    new_sparse_tensor = tf.sparse.reorder(new_sparse_tensor)
    return new_sparse_tensor


def sparse_gather_sub(x, sub_index, axis=0):
    is_sparse_matrix = isinstance(x, tfs.SparseMatrix)

    if is_sparse_matrix:
        sparse_tensor = x.to_sparse_tensor()
    else:
        sparse_tensor = x

    output_sparse_tensor = sparse_tensor_gather_sub(sparse_tensor, sub_index, axis)

    if is_sparse_matrix:
        return x.__class__.from_sparse_tensor(output_sparse_tensor)
    else:
        return output_sparse_tensor


def compute_num_or_size_splits(num_h_features, num_splits):

    if num_splits is None or num_splits == 1:
        num_or_size_splits = None
    elif num_h_features % num_splits == 0:
        num_or_size_splits = num_splits
    else:
        split_size = np.ceil(num_h_features / num_splits).astype(np.int64)

        num_pre_splits = np.floor(num_h_features / split_size).astype(np.int64)
        last_split_size = num_h_features % split_size

        num_or_size_splits = [split_size] * num_pre_splits + ([last_split_size] if last_split_size > 0 else [])

        if len(num_or_size_splits) != num_splits:
            raise Exception(
                "cannot split H of shape [None, {}] into {} matrices, please provide a valid num_splits"
                .format(num_h_features, num_splits))

    return num_or_size_splits



# coding=utf-8
import tensorflow as tf
from tf_geometric.nn import aggregate_neighbors


class MapReduceGNN(tf.keras.Model):

    def map(self, repeated_x, neighbor_x, edge_weight=None):
        pass

    def reduce(self, neighbor_msg, node_index, num_nodes=None):
        pass

    def update(self, x, reduced_neighbor_msg):
        pass

    def get_mapper(self):
        def mapper(repeated_x, neighbor_x, edge_weight=None):
            return self.map(repeated_x, neighbor_x, edge_weight)
        return mapper

    def get_reducer(self):
        def reducer(neighbor_msg, node_index, num_nodes=None):
            return self.reduce(neighbor_msg, node_index, num_nodes)
        return reducer

    def get_updater(self):
        def updater(x, reduced_neighbor_msg):
            return self.update(x, reduced_neighbor_msg)
        return updater

    def call(self, inputs, training=None, mask=None):
        x, edge_index, edge_weight = inputs
        return aggregate_neighbors(
            x,
            edge_index,
            edge_weight,
            self.get_mapper(),
            self.get_reducer(),
            self.get_updater()
        )